/**
 ******************************************************************************
 * STM32L4 Bootloader
 ******************************************************************************
 * @author Akos Pasztor
 * @file   bootloader.c
 * @brief  Bootloader implementation
 *	        This file contains the functions of the bootloader. The bootloader
 *	        implementation uses the official HAL library of ST.
 * @see    Please refer to README for detailed information.
 ******************************************************************************
 * Copyright (c) 2018 Akos Pasztor.                    https://akospasztor.com
 ******************************************************************************
 **/

#include "stm32f4xx.h"
#include "bootloader.h"
#include "main.h"
#include "stm32f4xx_hal.h"


#include "stm32f4xx_hal_flash_ex.h"

/* Base address of the Flash sectors */

uint32_t GetSector(uint32_t Address)
{
	uint32_t sector = 0;

	if((Address < ADDR_FLASH_SECTOR_1) && (Address >= ADDR_FLASH_SECTOR_0))
	{
		sector = FLASH_SECTOR_0;
	}
	else if((Address < ADDR_FLASH_SECTOR_2) && (Address >= ADDR_FLASH_SECTOR_1))
	{
		sector = FLASH_SECTOR_1;
	}
	else if((Address < ADDR_FLASH_SECTOR_3) && (Address >= ADDR_FLASH_SECTOR_2))
	{
		sector = FLASH_SECTOR_2;
	}
	else if((Address < ADDR_FLASH_SECTOR_4) && (Address >= ADDR_FLASH_SECTOR_3))
	{
		sector = FLASH_SECTOR_3;
	}
	else if((Address < ADDR_FLASH_SECTOR_5) && (Address >= ADDR_FLASH_SECTOR_4))
	{
		sector = FLASH_SECTOR_4;
	}
	else if((Address < ADDR_FLASH_SECTOR_6) && (Address >= ADDR_FLASH_SECTOR_5))
	{
		sector = FLASH_SECTOR_5;
	}
	else if((Address < ADDR_FLASH_SECTOR_7) && (Address >= ADDR_FLASH_SECTOR_6))
	{
		sector = FLASH_SECTOR_6;
	}
	else if((Address < ADDR_FLASH_SECTOR_8) && (Address >= ADDR_FLASH_SECTOR_7))
	{
		sector = FLASH_SECTOR_7;
	}
	else if((Address < ADDR_FLASH_SECTOR_9) && (Address >= ADDR_FLASH_SECTOR_8))
	{
		sector = FLASH_SECTOR_8;
	}
	else if((Address < ADDR_FLASH_SECTOR_10) && (Address >= ADDR_FLASH_SECTOR_9))
	{
		sector = FLASH_SECTOR_9;
	}
	else if((Address < ADDR_FLASH_SECTOR_11) && (Address >= ADDR_FLASH_SECTOR_10))
	{
		sector = FLASH_SECTOR_10;
	}
	else if((Address < ADDR_FLASH_SECTOR_12) && (Address >= ADDR_FLASH_SECTOR_11))
	{
		sector = FLASH_SECTOR_11;
	}
	else if((Address < ADDR_FLASH_SECTOR_13) && (Address >= ADDR_FLASH_SECTOR_12))
	{
		sector = FLASH_SECTOR_12;
	}
	else if((Address < ADDR_FLASH_SECTOR_14) && (Address >= ADDR_FLASH_SECTOR_13))
	{
		sector = FLASH_SECTOR_13;
	}
	else if((Address < ADDR_FLASH_SECTOR_15) && (Address >= ADDR_FLASH_SECTOR_14))
	{
		sector = FLASH_SECTOR_14;
	}
	else if((Address < ADDR_FLASH_SECTOR_16) && (Address >= ADDR_FLASH_SECTOR_15))
	{
		sector = FLASH_SECTOR_15;
	}
	else if((Address < ADDR_FLASH_SECTOR_17) && (Address >= ADDR_FLASH_SECTOR_16))
	{
		sector = FLASH_SECTOR_16;
	}
	else if((Address < ADDR_FLASH_SECTOR_18) && (Address >= ADDR_FLASH_SECTOR_17))
	{
		sector = FLASH_SECTOR_17;
	}
	else if((Address < ADDR_FLASH_SECTOR_19) && (Address >= ADDR_FLASH_SECTOR_18))
	{
		sector = FLASH_SECTOR_18;
	}
	else if((Address < ADDR_FLASH_SECTOR_20) && (Address >= ADDR_FLASH_SECTOR_19))
	{
		sector = FLASH_SECTOR_19;
	}
	else if((Address < ADDR_FLASH_SECTOR_21) && (Address >= ADDR_FLASH_SECTOR_20))
	{
		sector = FLASH_SECTOR_20;
	}
	else if((Address < ADDR_FLASH_SECTOR_22) && (Address >= ADDR_FLASH_SECTOR_21))
	{
		sector = FLASH_SECTOR_21;
	}
	else if((Address < ADDR_FLASH_SECTOR_23) && (Address >= ADDR_FLASH_SECTOR_22))
	{
		sector = FLASH_SECTOR_22;
	}
	else/*(Address < FLASH_END_ADDR) && (Address >= ADDR_FLASH_SECTOR_11))*/
	{
		sector = FLASH_SECTOR_23;
	}

	return sector;
}



/* Private variables ---------------------------------------------------------*/
static uint32_t flash_ptr = APP_ADDRESS;

#define FLASH_FLAG_ALL_ERRORS     (FLASH_FLAG_EOP | FLASH_FLAG_OPERR | FLASH_FLAG_WRPERR |\
		FLASH_FLAG_PGAERR | FLASH_FLAG_PGPERR|FLASH_FLAG_PGSERR)

/*** Initialize bootloader and flash ******************************************/
void Bootloader_Init(void)
{
	__HAL_RCC_SYSCFG_CLK_ENABLE();

	/* Clear flash flags */
	HAL_FLASH_Unlock();
	__HAL_FLASH_CLEAR_FLAG(FLASH_FLAG_ALL_ERRORS);
	HAL_FLASH_Lock();
}

/*** Erase flash **************************************************************/
uint8_t Bootloader_Erase(void)
{
	uint32_t StartSector = 0, EndSector = 0,i=0;
	FLASH_EraseInitTypeDef  pEraseInit;
	uint32_t PageError  = 0;
	HAL_StatusTypeDef       status = HAL_OK;
	status =  HAL_FLASH_Unlock();
	if( status == HAL_OK )
	{
		/* Get the number of the start and end sectors */
		StartSector = GetSector(APP_ADDRESS);
		EndSector = GetSector(END_ADDRESS);

		for (i = StartSector; i < EndSector; i++)
		{
			pEraseInit.TypeErase = FLASH_TYPEERASE_SECTORS;
			pEraseInit.VoltageRange = FLASH_VOLTAGE_RANGE_3;
			pEraseInit.Sector =  i;
			pEraseInit.NbSectors = 1;

			status =  HAL_FLASHEx_Erase(&pEraseInit, &PageError);
			if (status != HAL_OK)
			{
				return status;
			}
		}
	}
	HAL_FLASH_Lock();
	return (status == HAL_OK) ? BL_OK : BL_ERASE_ERROR;
}

/*** Begin flash programming **************************************************/
void Bootloader_FlashBegin(void)
{    
	/* Reset flash destination address */
	flash_ptr = APP_ADDRESS;
	/* Unlock flash */
	HAL_FLASH_Unlock();
}

/*** Program 64bit data into flash ********************************************/
uint8_t Bootloader_FlashNext(uint32_t data)
{
	if(HAL_FLASH_Program(FLASH_TYPEPROGRAM_WORD, flash_ptr, data) == HAL_OK)
		{
			/* Increment Flash destination address */
			if(*(uint32_t*)flash_ptr != data)
			{
				/* Flash content doesn't match source content */
				HAL_FLASH_Lock();
				return BL_WRITE_ERROR;
			}
			flash_ptr += 4;
		}
		else
		{
			/* Error occurred while writing data into Flash */
			HAL_FLASH_Lock();
			return HAL_ERROR;
		}

		return HAL_OK;
	return BL_OK;
}

/*** Finish flash programming *************************************************/
void Bootloader_FlashEnd(void)
{   
	/* Lock flash */
	HAL_FLASH_Lock();
}

/*** Get flash protection status **********************************************/


/*** Check if application fits into user flash ********************************/
uint8_t Bootloader_CheckSize(uint32_t appsize)
{
	//return ((FLASH_BASE + FLASH_DATA_SECTOR - APP_ADDRESS) >= appsize) ? BL_OK : BL_SIZE_ERROR;
}

/*** Verify checksum of application located in flash **************************/
uint8_t Bootloader_VerifyChecksum(void)
{
#if (USE_CHECKSUM)
	CRC_HandleTypeDef CrcHandle;
	volatile uint32_t calculatedCrc = 0;

	__HAL_RCC_CRC_CLK_ENABLE();
	CrcHandle.Instance = CRC;
	CrcHandle.Init.DefaultPolynomialUse    = DEFAULT_POLYNOMIAL_ENABLE;
	CrcHandle.Init.DefaultInitValueUse     = DEFAULT_INIT_VALUE_ENABLE;
	CrcHandle.Init.InputDataInversionMode  = CRC_INPUTDATA_INVERSION_NONE;
	CrcHandle.Init.OutputDataInversionMode = CRC_OUTPUTDATA_INVERSION_DISABLE;
	CrcHandle.InputDataFormat              = CRC_INPUTDATA_FORMAT_WORDS;
	if(HAL_CRC_Init(&CrcHandle) != HAL_OK)
	{
		return BL_CHKS_ERROR;
	}

	calculatedCrc = HAL_CRC_Calculate(&CrcHandle, (uint32_t*)APP_ADDRESS, APP_SIZE);

	__HAL_RCC_CRC_FORCE_RESET();
	__HAL_RCC_CRC_RELEASE_RESET();

	if( (*(uint32_t*)CRC_ADDRESS) == calculatedCrc )
	{
		return BL_OK;
	}
#endif
	return BL_CHKS_ERROR;
}

/*** Check for application in user flash **************************************/

/*** Jump to application ******************************************************/
void Bootloader_JumpToApplication(void)
{
	 uint32_t  JumpAddress = *(__IO uint32_t*)(APP_ADDRESS + 4);
	    pFunction Jump = (pFunction)JumpAddress;

	    HAL_RCC_DeInit();
	    HAL_DeInit();

	    SysTick->CTRL = 0;
	    SysTick->LOAD = 0;
	    SysTick->VAL  = 0;

	#if (SET_VECTOR_TABLE)
	    SCB->VTOR = APP_ADDRESS;
	#endif

	    __set_MSP(*(__IO uint32_t*)APP_ADDRESS);
	    Jump();
}

/*** Jump to System Memory (ST Bootloader) ************************************/
void Bootloader_JumpToSysMem(void)
{
	uint32_t  JumpAddress = *(__IO uint32_t*)(SYSMEM_ADDRESS + 4);
	    pFunction Jump = (pFunction)JumpAddress;

	    HAL_RCC_DeInit();
	    HAL_DeInit();

	    SysTick->CTRL = 0;
	    SysTick->LOAD = 0;
	    SysTick->VAL  = 0;

	    __HAL_RCC_SYSCFG_CLK_ENABLE();
	    __HAL_SYSCFG_REMAPMEMORY_SYSTEMFLASH();

	    __set_MSP(*(__IO uint32_t*)SYSMEM_ADDRESS);
	    Jump();

	    while(1);
}
